package it.polito.elite.doseWS.agency;

import java.io.*;
import java.util.*;
import java.text.DateFormat;
import java.text.SimpleDateFormat;

import org.w3c.dom.*;
import org.xml.sax.*;
import javax.xml.parsers.*;
import it.polito.elite.doseWS.include.*;
import it.polito.elite.doseWS.agency.parsers.*;
import it.polito.elite.doseWS.agency.messages.*;
import it.polito.elite.doseWS.agency.behaviours.*;
import jade.core.*;
import jade.lang.acl.*;
import javax.xml.namespace.QName;
import javax.xml.rpc.Call;
import javax.xml.rpc.Service;
import javax.xml.rpc.ServiceFactory;

/**
 * <p>Title: AnnotationAgent</p>
 * <p>Description: This employee agent is designed to create annotations with respect
 * of a given ontology on a textual fragment. It receives such fragments from
 * filter agents and maps the document to concept by means of sets of lexical entities
 * (M-Word) each related to a concept. The agent obtain these sets of m-word in a given language
 * by requiring them to a Synset Manager. After composing the annotations it sends them
 * to the AnnotationRepository WS. Afterwards it sends an annotation report to the
 * SynsetManager which is in charge of updating the M-Words weights and learning new ones.</p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: Politecnico di Torino</p>
 * @author Alessio Bosca
 */


public class AnnotationAgent extends DOSEAgent
{
  private static final long serialVersionUID = 1;
  static public final String SERVICE = "FragmentAnnotation";
  static private final int newTermMinOccur = 5;
  private Hashtable Mwords, conceptsAnnoProbability;
  private Stemmer stemmer;
  private String lang;
  public String serverAnnoWS, endpointAnnoWS;
  private Date expirationDate;
  private HashSet synsetManager, requests;
  private AID manager;
  private MsgParser msgParser;
  protected boolean busy;

  /**
   * Agent LifeCycle startup
   */
  public void setup()
  {
    //get initialization parameters
    Object params[] = this.getArguments();
    Location loc = (Location) params[0];
    this.endpointAnnoWS = (String) params[1];
    this.manager = (AID) params[2];
    //get AnnoWS server
    int indx = this.endpointAnnoWS.indexOf("://") + 3;
    indx = this.endpointAnnoWS.indexOf("/", indx);
    this.serverAnnoWS = this.endpointAnnoWS.substring(0, indx + 1);
    //We have no synsets loaded on start-up, since the first request, so lang =""
    this.lang = new String("");
    if (loc != null) {
      this.doMove(loc);
    }
    else {
      this.sendRemoteLog("!!Location of the destination container is null",manager);
      //Explicitly call afterMove when not migration the agent for debug pourposes
      //this.afterMove()
    }
  }

  /**
   * This method is invoked each time an agent is migrated to a new Location
   * and completes the agent setup phase.
   */
  public void afterMove()
  {
    //finishing setup...
    this.msgParser = new MsgParser();
    this.requests = new HashSet();
    this.synsetManager = new HashSet();
    this.stemmer = new Stemmer();
    this.Mwords = new Hashtable();
    //msgs from Agency Manager
    this.msgParser.register(IsAlive.class);
    this.msgParser.register(Die.class);
    //msgs from Filter Agent
    this.msgParser.register(AnnotationPreRequestMessage.class);
    this.msgParser.register(AnnotationRequestMessage.class);
    this.msgParser.register(SynsetResponseMessage.class);
    //behaviour
    this.addBehaviour(new ListenMessageBehaviour(this));
    this.addBehaviour(new GetAgentProviders(this, 600000, SynsetManager.SERVICE,
					    synsetManager));
    //register to Directory Facilitator
    this.register(AnnotationAgent.SERVICE);
  }

  /**
   * Overrides the same method of DOSEAgent. Handles an incoming ACLMessage
   * from the agency
   * @param msg Incoming Message
   */
  public void handleMessage(ACLMessage msg)
  {
    Message decodedMsg = msgParser.routeMessage(msg.getContent());
    if (decodedMsg instanceof IsAlive) {
      ImAlive content = new ImAlive(AnnotationAgent.SERVICE);
      ACLMessage reply = msg.createReply();
      reply.setContent(content.createMsg());
      this.send(reply);
    }
    else if (decodedMsg instanceof AnnotationPreRequestMessage) {
      AnnotationPreRequestMessage preReq = (AnnotationPreRequestMessage)
	  decodedMsg;
      if (lang.equals(preReq.getLanguage())) {
	ACLMessage reply = msg.createReply();
	AnnotationOfferMessage resp = new AnnotationOfferMessage(expirationDate,
	    this.getCurQueueSize() + 1);
	reply.setContent(resp.createMsg());
	this.send(reply);
	this.sendRemoteLog("Annotation Offer: queue=" + resp.getRunnings() +
			   " exp=" + resp.getDate().toString(), this.manager);
      }
    }
    else if (decodedMsg instanceof AnnotationRequestMessage) {
      AnnotationRequestMessage req = (AnnotationRequestMessage) decodedMsg;
      String reqLang = req.getLanguage();
      this.sendRemoteLog("Received request of " + req.getURL(), manager);
      if ( (!lang.equals(reqLang)) || (!this.isExpirationValid())) {
	//Request of new Synsets
	SynsetRequestMessage synReq = new SynsetRequestMessage(reqLang);
	ACLMessage synReqMsg = new ACLMessage(ACLMessage.REQUEST);
	synReqMsg.setContent(synReq.createMsg());
	Iterator it = this.synsetManager.iterator();
	AID sm = (AID) it.next();
	synReqMsg.addReceiver(sm);
	this.send(synReqMsg);
	//Receive the synset in a blocking way
	MessageTemplate mt = MessageTemplate.MatchSender(sm);
	ACLMessage synsMsg = this.blockingReceive(mt); //gestione timeout ed errore..
	this.sendRemoteLog("Received synsets... in " + reqLang, manager);
	SynsetResponseMessage srm = (SynsetResponseMessage)
	    SynsetResponseMessage.parse(synsMsg.getContent());
	//Load Synset into memory
	if (!this.acquireSynsets(srm.getContent())) {
	  this.sendRemoteLog(
	      "!!! Failure on loading Synsets.. Discarding document " +
	      req.getURL(), this.manager);
	  return;
	}
      }
      this.annotate(req.getText(), req.getURL());
    }
    else if (decodedMsg instanceof SynsetResponseMessage) {
      // The message can only be received after an explicit Synset request
      // following a annotation request. If others arrives they are just discarded
      this.sendRemoteLog("!!Received unexpected not requested synsets",manager);
    }
    else if (decodedMsg instanceof Die) {//Explicit Die request
      this.doDelete();
    }

  }

  /**
   * Tells wether the synset expiration valid is still valid or not
   * @param expirationDate The Expiration date of the current Synsets
   * @return boolean validiy value
   */
  private boolean isExpirationValid()
  {
    Date d = new Date(System.currentTimeMillis());
    if (d.compareTo(this.expirationDate) > 0) {
      return false;
    }
    else {
      return true;
    }
  }

  /**
   * Create the annotations and send them to AnnotationRepositoryWS. It further
   * compose a report message for the SynsetManager and another one for the agency
   * manager.
   * @param doc The text fragment to annotate
   * @param docURL the URl of the document to be annotated
   * @return true on success, false otherwise
   */
  public boolean annotate(String doc, String docURL)
  {
    Hashtable occurrences, results;
    occurrences = new Hashtable();
    results = new Hashtable();
    doc = this.resultingSynsets(doc, results, occurrences);
    this.composeAnnotations(results, docURL);
    HashSet cwords = filterNewWords(doc);
    //Send a report to the synset manager
    ACLMessage msg = new ACLMessage(ACLMessage.INFORM);
    LearningMessage lm = new LearningMessage();
    msg.setContent(lm.createMsg(occurrences, cwords, this.lang));
    Iterator it = this.synsetManager.iterator();
    msg.addReceiver( (AID) it.next());
    this.send(msg);
    //Send a report to agency manager
    msg = new ACLMessage(ACLMessage.INFORM);
    IndexReportMessage rm = new IndexReportMessage(docURL);
    msg.setContent(rm.createMsg());
    msg.addReceiver(manager);
    this.send(msg);
    return true;
  }

  /**
   * Retrieve a collection of the unrecognized terms whit the highest frequency.
   * It use an internal static threshold (newTermMinOccur) in order discriminate
   * between the relevant and the not relevant
   * @param doc The textual fragment containing the unrecognized words
   * @return the Collection of the filtered terms
   */
  private HashSet filterNewWords(String doc)
  {
    HashSet hs = new HashSet();
    String[] splitted = doc.split("[ ]+");
    Arrays.sort(splitted);
    String old = splitted[0];
    int cont = 0;
    for (int i = 0; i < splitted.length; i++) {
      if (splitted[i].equals(old)) {
	cont++;
      }
      else {
	if (cont > AnnotationAgent.newTermMinOccur) {
	  hs.add(new LearningPair(old, cont));
	}
	old = splitted[i];
	cont = 0;
      }
    }
    return hs;
  }

  /**
   * The actual method for annotation composition
   * @param t Hashtable of the M-Words collection
   * @param URL the URL of the document, used as a term of an Annotation
   */
  private void composeAnnotations(Hashtable t, String URL)
  {
    Enumeration e = t.keys();
    while (e.hasMoreElements()) {
      //Retrieve weighted mwords for each concept present in the result hashtable
      //and compute the annotation
      String concept = (String) e.nextElement();
      Vector v = (Vector) t.get(concept);
      float annoWeight = 1;
      for (int i = 0; i < v.size(); i++) {
	annoWeight *= (1 - ( (RelationToConcept) v.elementAt(i)).weight);
      }
      annoWeight = (1 - annoWeight) *
	  ( (Float)this.conceptsAnnoProbability.get(concept)).floatValue();

      this.sendRemoteLog("Send Annotation: " + URL + " " + concept + " " +
			 annoWeight + " " +
			 lang, this.manager);

      //Send to Annotation WS
      try {
	//creates a new QName identifier for the annotation service "http://localhost:8080/"
	QName op = new QName(this.serverAnnoWS, "addAnnotation");

	//creating service factory
	ServiceFactory factory = ServiceFactory.newInstance();
	//defining the service
	Service service = factory.createService(new QName(
	    "AnnoRepositoryWSService"));
	//defining the service port
	QName port = new QName("AnnoRepositoryWS");
	//creating the call object
	Call call = service.createCall(port);
	//sets the endpoint for the call
	call.setTargetEndpointAddress(this.endpointAnnoWS);

	//Setting up the params
	Object params[] = new Object[6];
	params[0] = URL;
	params[1] = "/html/body"; //xpath
	params[2] = concept;
	params[3] = new Double(annoWeight);
	params[4] = "default";
	params[5] = "default";

	//sets the operation to be required by the call
	call.setOperationName(op);
	Boolean ret = (Boolean) call.invoke(params);
      }
      catch (Exception ex) {
	this.sendRemoteLog("!! Calling AnnoWS: " + ex.toString(), this.manager);
      }

    } //end while

  }

  /**
   * Internal method for document analisys and M-Word frequency retrieval
   * @param doc The textual fragment
   * @param results The M-Words occurring within the document
   * @param occurrences Their Occurencies
   * @return the textual fragment pruned of the M-Words
   */
  private String resultingSynsets(String doc, Hashtable results,
				  Hashtable occurrences)
  {
    Enumeration enumMwords = this.Mwords.keys();
    //Searching the document for the mword in the synset.
    while (enumMwords.hasMoreElements()) {
      String mWord = (String) enumMwords.nextElement();
      String stemmed = this.stemmer.stem(mWord);
      if (doc.indexOf(stemmed) >= 0) {

	//get the number of occurrences
	int occ = 1;
	int index = 0;
	while ( (index = doc.indexOf(stemmed, index + 1)) >= 0) {
	  occ++;
	}
	//remove from the text
	doc = doc.replaceAll(stemmed, "");

	//Get all the concepts related to the just found mword and for each concept
	//properly update the result Hashtable conceptURI -> (mword,weight)+
	//and occurrences Hastable conceptURI -> (mword,occ)+
	Enumeration e = ( (Vector) Mwords.get(mWord)).elements();
	while (e.hasMoreElements()) {

	  //weighted mword relation to a concept
	  RelationToConcept r = (RelationToConcept) e.nextElement();
	  Vector relToConceptVector, learningPairVector;

	  //adding a new RelationToConcept to results hashtable
	  if (results.containsKey(r.uriConcept)) {
	    relToConceptVector = (Vector) results.get(r.uriConcept);
	  }
	  else {
	    relToConceptVector = new Vector(5, 2);
	  }
	  relToConceptVector.add(new RelationToConcept(mWord, r.weight));
	  results.put(r.uriConcept, relToConceptVector);
	  //adding a new LearninPair to occurrences hashtable
	  if (occurrences.containsKey(r.uriConcept)) {
	    learningPairVector = (Vector) occurrences.get(r.uriConcept);
	  }
	  else {
	    learningPairVector = new Vector(5, 2);
	  }
	  learningPairVector.add(new LearningPair(mWord, occ));
	  occurrences.put(r.uriConcept, learningPairVector);
	} //end while
      } //if contains

    } //while enumMwords...
    return doc;
  }

  /**
   * Reads in memory the a lookUpTable from Sysnsets
   * if its expirationDate is still valid..
   * @param f XML file containing the Synsets
   * @return true on success, false otherwise
   */
  private boolean acquireSynsets(String content)
  {
    try {
      Document doc;
      DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
      DocumentBuilder builder = factory.newDocumentBuilder();
      //Reader for the content
      StringReader reader = new StringReader(content);
      //InputSource for the content
      InputSource source = new InputSource(reader);

      doc = builder.parse(source);
      Element elemRoot = doc.getDocumentElement();

      String expiration = elemRoot.getAttribute("expiration");
      //Expiration Date
      SimpleDateFormat df = (SimpleDateFormat) DateFormat.getDateTimeInstance();
      df.applyPattern("yyyy:MM:dd:hh:mm:ss:S:Z");
      Date myDate = df.parse(expiration);
      this.expirationDate = myDate;
      this.Mwords = new Hashtable();

      org.w3c.dom.NodeList concepts = elemRoot.getElementsByTagName("SYNSET");
      conceptsAnnoProbability = new Hashtable(concepts.getLength());
      for (int i = 0; i < concepts.getLength(); i++) {
	Element concept = (Element) concepts.item(i);
	String uriConcept = concept.getAttribute("concept");
	String annoProbStr = concept.getAttribute("annoProbability");
	conceptsAnnoProbability.put(uriConcept, new Float(annoProbStr));
	if (i == 0) {
	  //Language
	  this.lang = ( (Element) concept.getElementsByTagName("SYNSETENTRY").
		       item(0)).
	      getAttribute("lang");
	  this.stemmer.initStemmer(Lang.getExtended(lang));
	}
	org.w3c.dom.NodeList mwords = concept.getElementsByTagName("MWORD");
	for (int j = 0; j < mwords.getLength(); j++) {
	  Element word = (Element) mwords.item(j);
	  String val = word.getAttribute("weight");
	  try {
	    float weight = Float.parseFloat(val);
	    String term = ( (org.w3c.dom.Text) word.getFirstChild()).getData();
	    if (!Mwords.containsKey(term)) {
	      Vector relations = new Vector(1, 1);
	      relations.add(new RelationToConcept(uriConcept, weight));
	      Mwords.put(term, relations);
	    }
	    else {
	      Vector relations = (Vector) Mwords.remove(term);
	      relations.add(new RelationToConcept(uriConcept, weight));
	      Mwords.put(term, relations);
	    }
	  }
	  catch (Exception ex) {
	    this.sendRemoteLog("!! AcquireSynsets -1: " + ex.toString(),
			       manager);
	    return false;
	  }
	} //end cycle on terms
      } //end cycle on concepts
    }
    catch (Exception exc) {
      this.sendRemoteLog("!! AcquireSynsets -2: " + exc.toString(), manager);
      return false;
    }
    return true;
  }

}