package it.polito.elite.doseWS.agency;

import java.util.*;
import javax.xml.parsers.*;

import org.htmlparser.*;
import org.htmlparser.filters.*;
import org.htmlparser.parserapplications.*;
import org.htmlparser.tags.*;
import org.htmlparser.util.*;
import org.w3c.dom.*;
import it.polito.elite.doseWS.agency.behaviours.*;
import it.polito.elite.doseWS.agency.messages.*;
import it.polito.elite.doseWS.agency.parsers.*;
import it.polito.elite.doseWS.include.*;
import jade.core.*;
import jade.lang.acl.*;
import sun.text.Normalizer;

/**
 * <p>Title: FilterAgent</p>
 * <p>Description: This employee agent is designed to filter a given textual fragment
 * formatting it to according to XHTML paradigm, detecting its language and removing stopwords.
 * Once detected the language it sends a preRequest to every AnnotationAgent specifying the
 * language of the document and then sends the actual request to the best offer or to
 * a random AnnotationAgent.</p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: Politecnico di Torino</p>
 * @author Alessio Bosca
 */

public class FilterAgent extends DOSEAgent
{
  private static final long serialVersionUID = 2;
  static public final String SERVICE = "HTML_DocumentFiltering";

  //dimensions of a generic window used to find out the main frame within framesets
  private static final int HEIGHT = 768, WIDTH = 1024;
  //Number of the same language's stopwords to be retrieve to 'recognize' a language
  private static final int LANG_FOUND = 4;
  private int LangDetectMax = 0;

  private String doc, stopWordsFileURL;
  //It contains a set of stopword foe each supported language
  private Hashtable stopDictionary;
  //Snowball Stemmer Engine
  private Stemmer stemmer;
  //The message parser
  private MsgParser msgParser;
  //Aid of annotator agents and Collection of URL to be filtered
  public HashSet requests, annotators;
  //Manager AID and references of the actual best annotation offer
  private AID bestOffer, manager;
  private int bestOfferRunnings;
  private Date bestOfferExpiration;

  /**
   * Agent LifeCycle startup
   */
  public void setup()
  {
    //get initialization parameters
    Object params[] = this.getArguments();
    Location loc = (Location) params[0];
    stopWordsFileURL = (String) params[1];
    manager = (AID) params[2];
    if (loc != null) {
      this.doMove(loc);
    }
    else {
      this.sendRemoteLog("!!Location of the destination container is null",
			 manager);
    }
    //Explicitly call afterMove when not migration the agent for debug pourposes
    //this.afterMove()

  }

  /**
   * This method is invoked each time an agent is migrated to a new Location
   * and completes the agent setup phase.
   */
  public void afterMove()
  {
    //finishing setup...
    this.acquireStopWords();
    this.msgParser = new MsgParser();
    this.requests = new HashSet();
    this.annotators = new HashSet();
    this.stemmer = new Stemmer();
    //msgs from Agency Manager
    this.msgParser.register(IsAlive.class);
    this.msgParser.register(IndexRequestMessage.class);
    this.msgParser.register(Die.class);
    //msgs from Annotator Agent
    this.msgParser.register(AnnotationOfferMessage.class);
    this.msgParser.register(AnnotationAcceptanceMessage.class);
    //behaviour
    this.addBehaviour(new ListenMessageBehaviour(this));
    this.addBehaviour(new CheckFilteringRequests(this));
    this.addBehaviour(new GetAgentProviders(this, 30000,
					    AnnotationAgent.SERVICE,
					    annotators));
    //register to Directory Facilitator
    this.register(FilterAgent.SERVICE);

  }

  /**
   * Overrides the same method of DOSEAgent. Handles an incoming ACLMessage
   * from the agency
   * @param msg Incoming Message
   */
  public void handleMessage(ACLMessage msg)
  {
    Message decodedMsg = msgParser.routeMessage(msg.getContent());
    if (decodedMsg instanceof IsAlive) {
      ImAlive content = new ImAlive(FilterAgent.SERVICE);
      ACLMessage reply = msg.createReply();
      reply.setContent(content.createMsg());
      this.send(reply);
    }
    else if (decodedMsg instanceof IndexRequestMessage) {
      IndexRequestMessage req = (IndexRequestMessage) decodedMsg;
      requests.add(req.getRequestedURL());
      this.sendRemoteLog("ho ricevuto index request: " + req.getRequestedURL(),
			 this.manager);
    }
    else if (decodedMsg instanceof AnnotationOfferMessage) {
      AnnotationOfferMessage off = (AnnotationOfferMessage) decodedMsg;
      if ( (this.bestOffer == null) ||
	  ( (off.getRunnings() <= this.bestOfferRunnings) &&
	   (this.bestOfferExpiration.compareTo(off.getDate()) < 0))) {
	//Update best offer
	bestOffer = msg.getSender();
	this.bestOfferRunnings = off.getRunnings();
	this.bestOfferExpiration = off.getDate();
      }
    }
    else if (decodedMsg instanceof AnnotationAcceptanceMessage) {
      //not yet implemented
    }
    else if (decodedMsg instanceof Die) {
      this.doDelete();
    }
  }

  /**
   * The core method for the various filtering phases management: language detecting,
   * PreRequest message, nouisances and stopwords removal, Stemming, Request message
   * @param URL
   */
  public void filter(String URL)
  {
    this.getText(getRealContentURL(URL));
    this.removeNuisances();
    String lang = detectLanguage();
    //preprequest
    this.sendPreRequest(lang);
    if (!this.stemmer.getLanguage().equals(lang)) {
      this.stemmer.initStemmer(Lang.getExtended(lang));
    }
    this.sendRemoteLog("sent prerequest of " + lang, manager);
    this.removeStopWords(lang);
    this.stemDoc();
    this.sendRequest(lang, URL);
    this.sendRemoteLog("sent Request: " + URL, manager);
  }

  /**
       * Sends a PreRequest message to each Annotation Agent specifying the language
   * used.
   * @param lang Language of the document
   */
  private void sendPreRequest(String lang)
  {
    AnnotationPreRequestMessage preReq = new AnnotationPreRequestMessage(lang);
    ACLMessage msg = new ACLMessage(ACLMessage.QUERY_REF);
    msg.setContent(preReq.createMsg());
    Iterator it = this.annotators.iterator();
    while (it.hasNext()) {
      msg.addReceiver( (jade.core.AID) it.next());
    }
    //Clear the previous best offer
    this.send(msg);
    this.bestOffer = null;
    this.bestOfferRunnings = 1000; //set an arbitrary high value
    this.bestOfferExpiration = new Date(System.currentTimeMillis()); //set expiration = now
  }

  /**
   * Send an annotation request in a given language for a textual fragment form the
   * given document
   * @param lang Langiage of the document
   * @param url URL of the document
   */
  private void sendRequest(String lang, String url)
  {
    try {
      AnnotationRequestMessage req = new AnnotationRequestMessage(doc, lang,
	  url);
      ACLMessage msg = new ACLMessage(ACLMessage.QUERY_REF);
      msg.setContent(req.createMsg());
      if (this.bestOffer != null) { //There are offers...
	msg.addReceiver(this.bestOffer);
      }
      else { //No offer pick a random annotator
	java.util.Random r = new java.util.Random(System.currentTimeMillis());
	AID[] aids = new AID[1];
	aids = (AID[])this.annotators.toArray(aids);
	int k = r.nextInt(aids.length);
	msg.addReceiver(aids[k]);
      }
      this.send(msg);
    }
    catch (Exception ex) {
      this.sendRemoteLog("!! SendRequest: " + ex.toString(), this.manager);
    }
  }

  /**
   * Stems the document
   */
  private void stemDoc()
  {
    StringTokenizer st = new StringTokenizer(doc);
    StringBuffer b = new StringBuffer();
    while (st.hasMoreTokens()) {
      String s = st.nextToken();
      s = stemmer.stem(s);
      b = b.append(" ").append(s);
    }
    doc = b.toString();
  }

  /**
   * Detects the language
   * @return The language of the document as a full lenght english string (e.g. french, english..)
   */
  private String detectLanguage()
  {
    String[] languages = new String[1];
    try {
      languages = (String[]) stopDictionary.keySet().toArray(languages);
    }
    catch (Exception ex) {
      this.sendRemoteLog("!! DetectLang: " + ex.toString(), this.manager);
    }
    if (languages.length < 1) {
      return null;
    }

    int[] coreWords = new int[languages.length];
    int langIndex = -1;
    Arrays.fill(coreWords, 0);
    boolean trovata = false;
    for (int i = 0; i < this.LangDetectMax && !trovata; i++) {

      for (int j = 0; j < languages.length; j++) {
	String stopword = (String) ( (Vector) stopDictionary.get(languages[j])).
	    elementAt(i);
	if (doc.indexOf(" " + stopword + " ") >= 0) {
	  coreWords[j]++;
	  if (coreWords[j] >= FilterAgent.LANG_FOUND) {
	    trovata = true;
	    langIndex = j;
	    break;
	  }
	}
      } //for j

    } //for i
    if (!trovata) {
      return null;
    }
    else {
      return languages[langIndex];
    }
  }

  /**
   * Load the Stopwords from an XML file from a web server
   * @return true on success, false otherwise
   */
  private boolean acquireStopWords()
  {
    try {
      Document doc;
      DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
      DocumentBuilder builder = factory.newDocumentBuilder();
      doc = builder.parse(this.stopWordsFileURL);
      Element elemRoot = doc.getDocumentElement();

      stopDictionary = new Hashtable();

      org.w3c.dom.NodeList dictionaries = elemRoot.getElementsByTagName(
	  "dictionary");
      stopDictionary = new Hashtable(dictionaries.getLength());

      for (int i = 0; i < dictionaries.getLength(); i++) {
	Element dict = (Element) dictionaries.item(i);
	String language = dict.getAttribute("language");
	String sizeStr = dict.getAttribute("size");
	int size = Integer.parseInt(sizeStr);
	Vector v = new Vector(size);

	org.w3c.dom.NodeList core = dict.getElementsByTagName("core");
	if (core.getLength() > this.LangDetectMax) {
	  this.LangDetectMax = core.getLength();
	}
	for (int j = 0; j < core.getLength(); j++) {
	  Element word = (Element) core.item(j);
	  String term = ( (org.w3c.dom.Text) word.getFirstChild()).getData();
	  v.add(term);
	} //end cycle on core

	org.w3c.dom.NodeList stop = dict.getElementsByTagName("stop");
	for (int j = 0; j < stop.getLength(); j++) {
	  Element word = (Element) stop.item(j);
	  String term = ( (org.w3c.dom.Text) word.getFirstChild()).getData();
	  v.add(term);
	} //end cycle on stop

	stopDictionary.put(language, v);

      } //end cycle on languages
    }
    catch (Exception exc) {
      this.sendRemoteLog("!! AcquireStopWords: " + exc.toString(), this.manager);
      return false;
    }
    return true;
  }

  /**
   * Retrieves the URL of the data, testing wether resources are sunken
   * in a frameset
   * @param URL to be analyzed
   * @return URL containing the real data
   */
  private String getRealContentURL(String URL)
  {
    org.htmlparser.util.NodeList Nlist = null;
    try {
      Parser parser = new Parser(URL);
      Nlist = parser.extractAllNodesThatMatch(new TagNameFilter("FRAMESET"));
    }
    catch (ParserException ex) {
      return URL;
    }

    if (Nlist.size() > 0) {
      FrameSetTag fs = (FrameSetTag) Nlist.elementAt(0);
      FrameTag f = null;
      while (f == null) {
	int nRow, nCol, totCols, son;

	String rows = fs.getAttribute("rows");
	String cols = fs.getAttribute("cols");
	nRow = getBiggerStripe(rows, FilterAgent.HEIGHT);
	nCol = getBiggerStripe(cols, FilterAgent.WIDTH);

	//Calcolo quante colonne ci sono
	if (cols != null) {
	  StringTokenizer st = new StringTokenizer(cols, ",");
	  totCols = st.countTokens();
	}
	else {
	  totCols = 1;
	}
	son = ( (nRow - 1) * totCols) + nCol - 1; //-1 xk 0 based

	Nlist = fs.getFrames();
	Nlist = Nlist.extractAllNodesThatMatch(
	    new OrFilter(new TagNameFilter("FRAME"),
			 new TagNameFilter("FRAMESET")));

	TagNameFilter tf = new TagNameFilter("FRAMESET");

	if (tf.accept(Nlist.elementAt(son))) { //frameset annidato, continuo a scendere
	  fs = (FrameSetTag) Nlist.elementAt(son);
	}
	else { //trovato frame principale
	  f = (FrameTag) Nlist.elementAt(son);
	}
      } //end while
      return f.getFrameLocation();
    }
    else {
      return URL;
    }
  }

  /**
   * Retrives the text from URL
   * @param URL Location of the document
   * @return the Document text
   */
  private void getText(String URL)
  {
    try {
      StringExtractor n = new StringExtractor(URL);
      doc = " " + n.extractStrings(false) + " ";
    }
    catch (ParserException ex) {
      this.sendRemoteLog("!! getText: " + ex.toString(), this.manager);
      doc = null;
    }
  }

  /**
   * Removes not XHTML compliant charachters from the document
   */
  private void removeNuisances()
  {
    doc = Normalizer.normalize(doc, Normalizer.DECOMP, 0);
    doc = doc.replaceAll("[^a-zA-Z]", " ");
    doc = doc.toLowerCase();
  }

  /**
   * remove Stop Words from the document text
   * @param doc teh document text
   * @param stopWords an array of stop words
   * @return the cleaned document
   */
  private void removeStopWords(String language)
  {
    Vector v = (Vector) stopDictionary.get(language);
    for (int i = 0; i < v.size(); i++) {
      doc = doc.replaceAll(" " + (String) v.elementAt(i) + " ", " ");
    }
  }

  /**
   *
   * @param set String representing the value for rows or cols
   * attribute in the frameset
   * @param max a default value for the lenght (cols) or the height (rows)
   * of the page. Such values are defined in HEIGHT and WIDTH constants
   * presents in the page
   * @return the positional order of the biggest stripe (row or column).
   * If the String set is null, default value of 1 is returned
   */
  private int getBiggerStripe(String set, int max)
  {
    if (set == null) {
      return 1;
    }

    StringTokenizer st = new StringTokenizer(set, ",");
    int numbers[] = new int[st.countTokens()];
    int Stars = 0, tot = 0, cont = 0;

    //conto quanti asterischi ci sono
    while ( (tot = set.indexOf('*', tot)) > 0) {
      Stars++;
    }

    //converto tutte % in pixel asumendo dimensioni di default
    //converto * in valori negativi pari al moltiplicatore
    //e calcolo somma parziale
    while (st.hasMoreTokens()) {
      String s = st.nextToken();
      if (s.indexOf('*') > 0) {
	s = s.replace('*', ' ');
	s = s.trim();
	try {
	  numbers[cont] = Integer.parseInt("-" + s);
	}
	catch (NumberFormatException ex) {
	  numbers[cont] = -1;
	}
      }
      else if (s.indexOf('%') > 0) {
	s = s.replace('%', ' ');
	s = s.trim();
	try {
	  numbers[cont] = Integer.parseInt(s);
	}
	catch (NumberFormatException ex) {
	  numbers[cont] = 1;
	}
	numbers[cont] = (numbers[cont] * max) / 100;

	tot += numbers[cont];
      }
      else {
	s = s.trim();
	try {
	  numbers[cont] = Integer.parseInt(s);
	}
	catch (NumberFormatException ex) {
	  numbers[cont] = 1;
	}

	tot += numbers[cont];
      }

      cont++;
    } //end while

    if (Stars > 0) {
      int resto = ( -1) * ( (max - tot) / Stars);
      for (int i = 0; i < numbers.length; i++) {
	if (numbers[i] < 0) {
	  numbers[i] *= resto;
	}
      }
    }

    int imaxim = 0;
    for (int i = 0; i < numbers.length; i++) {
      if (numbers[i] > numbers[imaxim]) {
	imaxim = i;
      }
    } //end for

    return imaxim + 1; //+1 in modo da essere 1 based
  }

}