/*
 * SpectrumImpl.java
 *
 * Created on July 5, 2004, 4:34 PM
 */

package it.polito.elite.doseWS.include;

import java.util.Hashtable;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Iterator;
import java.util.StringTokenizer;
import java.lang.Double;
import java.io.Serializable;

/**
 *
 * A java class for representing conceptual spectra, such class encapsulates all information
 * required in the definition of a conceptual spectra providing methods for evaluating the distance
 * between spectra according to a similarity value defined as:<br>
 * <center>a*formSimilarity+(1-a)*energySimilarity</center>
 * 
 * @author  <a href="mailto:dario.bonino@polito.it">Dario Bonino</a>
 * <a href="mailto:alessio.bosca@polito.it">Alessio Bosca</a>
 * <a href="mailto:federico.pescarmona@polito.it">Federico Pescarmona</a>
 * <a href="mailto:michele.debandi@polito.it">Michele Debandi</a>
 * 
 * @see The SpectrumImplTest for the corresponding test class.
 * @version 1.0
 * @since 11/2004
 */
public class SpectrumImpl implements Cloneable,Serializable,Spectrum
{
	private static double alpha = 0.9;
	private static double qu = 1.3;
	//Instance fields
	private String fragmentURI;
	private String fragmentXPath;
	private String author;
	private String type;
	private Hashtable conceptsAndWeights;
	
	/** 
	 * Constructor
	 * 
	 * @param fragmentURI the URI of the resource that this spectrum represent
	 * @param fragmentXPath the XPath identifying the specific substructure of the
	 * resource to which this spectrum is referred 
	 * */
	public SpectrumImpl(String fragmentURI, String fragmentXPath, String author, String type, Hashtable conceptsAndWeights)
	{
		this.fragmentURI = fragmentURI;
		this.fragmentXPath = fragmentXPath;
		this.author = author;
		this.type = type;
		this.conceptsAndWeights = this.copySpectrum(conceptsAndWeights);
	}
	
	/**
	 * @return the URI of the resource to which this spectrum is referred.
	 */
	public String getFragmentURI()
	{
		return this.fragmentURI;
	}
	
	/**
	 * @return the XPath of the resource to which this spectrum is referred
	 */
	public String getFragmentXPath()
	{
		return this.fragmentXPath;
	}
	/**
	 * @return the name of the person or of the machine who generated this spectrum
	 */
	public String getAuthor()
	{
		return this.author;
	}
	/**
	 * @return the type of author
	 */
	public String getType()
	{
		return this.type;
	}
	
	/**
	 * A method for hardcopying hashtables
	 * @param spectrum the internal hashtable of the spectrum to be copied
	 * @return the hardcopy of the given hashtable
	 */
	private Hashtable copySpectrum(Hashtable spectrum)
	{
		Hashtable spectrumNew = new Hashtable();
		Enumeration keys = spectrum.keys();
		while (keys.hasMoreElements())
		{
			String key = (String) keys.nextElement();
			spectrumNew.put(key,spectrum.get(key));
		}
		return spectrumNew;
	}
	/**
	 * @return the concepts contained in this conceptual spectrum together with their weights, in an
	 * Hashtable object.
	 */
	public Hashtable getConcepts()
	{
		return this.copySpectrum(this.conceptsAndWeights);
	}
	
	/**
	 * A method for evaluating the shape distance between two spectra, according to the
	 * cosine similarity value 
	 * 
	 * @param sa the first spectrum
	 * @param sb the second spectrum
	 * @return the distance between the two spectra expressed as a double in the range between 0 and 1.
	 */
	private double shapeDist(Spectrum sa, Spectrum sb)
	{
		/*evaluates the distance between the two spectra shapes as the cosine of
		 *the angle between them
		 */
		
		//get all components of the spectrum sa
		Hashtable sA = sa.getConcepts();
		//get all components of the spectrum sb
		Hashtable sB = sb.getConcepts();
		
		//creates a common topic set with no duplicates
		HashSet topics = new HashSet(sA.keySet());
		topics.addAll(sB.keySet());
		
		//compute the cosine of the angle between spectras 
		Iterator topicsIter = topics.iterator();
		double ab=0.0,a2=0.0,b2=0.0;
		while (topicsIter.hasNext())
		{
			double ai=0.0,bi=0.0;
			String topic = (String) topicsIter.next();
			
			//extracting sa component
			Double component = (Double)sA.get(topic);
			if(component != null)
				ai = component.doubleValue();
						
			//extracting sb component
			component = (Double)sB.get(topic);
			if(component != null)
				bi = component.doubleValue();
			
			//computing scalar product and modules
			ab+=ai*bi;
			a2+=(ai*ai);
			b2+=(bi*bi);
		}
		
		//consine of the angle between spectra vectors
		// sqrt(ab)=sqrt(a)*sqrt(b)
		double cosine = ab/Math.sqrt(a2*b2);
		
		return cosine;
	}
	
	/**
	 * A method for evaluating the shape distance between two spectra, according to the
	 * energy similarity value 
	 * @param sa the first spectrum
	 * @param sb the second spectrum
	 * @return the energy similarity value as a double
	 */
	private double energyDist(Spectrum sa, Spectrum sb)
	{
		/*evaluates the distance between the two spectra energies
		 *as the distance between spectra vectors
		 */
		
		//get all components of the spectrum sa
		Hashtable sA = sa.getConcepts();
		//get all components of the spectrum sb
		Hashtable sB = sb.getConcepts();
		
		//creates a common topic set with no duplicates
		HashSet topics = new HashSet(sA.keySet());
		topics.addAll(sB.keySet());
		
		//compute the energy distance between spectras 
		Iterator topicsIter = topics.iterator();
		double dist=0.0;
		while (topicsIter.hasNext())
		{
			double ai=0.0,bi=0.0;
			String topic = (String) topicsIter.next();
			
			//extracting sa component
			Double component = (Double)sA.get(topic);
			if(component != null)
				ai = component.doubleValue();
						
			//extracting sb component
			component = (Double)sB.get(topic);
			if(component != null)
				bi = component.doubleValue();
			
			//computing distance (least square method)
			dist+=(ai-bi)*(ai-bi);
		}
		
		//energy
		double energy = Math.sqrt(dist);
		
		return energy;
	}
	/**
	 * A method for evaluating the similarity between two spectra according to the similarity definition
	 * given in the class description
	 * @param sa the first spectrum
	 * @param sb the second spectrum
	 * @return the similarity of the two spectra as a double in the range between 0 and 1
	 */
	public double dist(Spectrum sa,Spectrum sb)
	{
		double shape = this.shapeDist(sa,sb);
		double energy = this.energyDist(sa,sb);
		double distance = SpectrumImpl.alpha*shape+(1-SpectrumImpl.alpha)*Math.pow(SpectrumImpl.qu,-1.0*energy);
		return distance;
	}
	
	/**
	 * A static setter for the tune-up parameters in the similarity function
	 * @param alpha the first parameter (a)
	 * @param qu the second parameter (b)
	 */
	public static void setDistParameters(double alpha,double qu)
	{
		SpectrumImpl.alpha = alpha;
		SpectrumImpl.qu = qu;
	}
		
	/**
	 * The cloner for this class
	 */
	public Object clone() throws CloneNotSupportedException
	{
		SpectrumImpl clone = new SpectrumImpl(this.fragmentURI,this.fragmentXPath,this.author,this.type,this.conceptsAndWeights);
		return (Object)clone;
	}
	/**
	 * The serializer for this class
	 * @return this spectrum object serialized into a String
	 */
	public String serialize()
	{
		String serialized = new String();
		String sep = ";";
		serialized = serialized+this.author+sep;
		serialized = serialized+this.fragmentURI+sep;
		serialized = serialized+this.fragmentXPath+sep;
		serialized = serialized+this.type+sep;
		
		Enumeration keys = this.conceptsAndWeights.keys();
		while (keys.hasMoreElements())
		{
			String key = (String)keys.nextElement();
			Double value = (Double)this.conceptsAndWeights.get(key);
			serialized = serialized+key+"="+value+sep;
		}
		return serialized;
	}
	/**
	 * The static deserializer for this class
	 * @param serialized the serialized spectrum
	 * @return a SpectrumImpl object corresponding to the serialized object
	 * contained in "serialized"
	 */
	public static SpectrumImpl deSerialize(String serialized)
	{
		//separator char
		String sep = ";";
		//internal fields
		String author;
		String fragmentURI;
		String fragmentXPath;
		String type;
		Hashtable conceptsAndWeights = new Hashtable();
		// initializinf tokenizer
		StringTokenizer strTok = new StringTokenizer(serialized,sep);
		
		//extracting fields
		author = new String(strTok.nextToken());
		fragmentURI = new String(strTok.nextToken());
		fragmentXPath = new String(strTok.nextToken());
		type = new String(strTok.nextToken());
		
		//extracting hashtable
		while (strTok.hasMoreTokens())
		{
			String entry = strTok.nextToken();
			int eqIndex = entry.indexOf('=');
			String key = entry.substring(0,eqIndex);
			Double value = new Double(Double.parseDouble(entry.substring(eqIndex+1)));
			conceptsAndWeights.put(key,value);
		}
		return new SpectrumImpl(fragmentURI,fragmentXPath,author,type,conceptsAndWeights);
	}
	
	/**
	 * Redefinition of the equals function for the class SpectrumImpl
	 */
	public boolean equals(Object s)
	{
		if(((this.fragmentURI).equals(((SpectrumImpl)s).fragmentURI))
				&&
				((this.fragmentXPath).equals(((SpectrumImpl)s).fragmentXPath)))
			return true;
		else
			return false;
	}
}
