/*
 * DBWrapper.java
 *
 * Created on July 5, 2004, 5:34 PM
 */
package it.polito.elite.doseWS.wrappers;
import it.polito.elite.doseWS.include.*;
import java.util.*;
import java.lang.Integer;
/**
 * Declared here are the methods defined for accessing the underlying
 * database structure for H-DOSE.
 *
 * See DBWRapperImpl for more informations.
 * @author Michele Debandi
 * @version 1
 */
public interface DBWrapper {
    /**
     * Insert an annotaions' block gotten from a spectrum
     * @param spectrum The Spectrum that will be inserted into the DB
     * @return A positive integer (number of inserted rows) for succesfull completion,
     * otherwise a negative integer
     *
     * @throws Exception For database errors
     */
    public int insertSpectrum(Spectrum spectrum) throws Exception ;
    /**
     * Inserts a block of refined annotations
     * @param spectrum The Spectrum that will be inserted into the DB
     * @return A positive integer (number of inserted rows) for succesfull completion,
     * otherwise a negative integer
     *
     * @throws Exception For database errors
     */
    public int insertXSpectrum   (Spectrum spectrum) throws Exception;
    /**
     * restituisce il blocco di annotazioni riferite ad un frammento
     * (autore e tipo sono facoltativi)
     * @return ritorna uno spetto
     * @param uri uri del frammento
     * @param xPath xpath del frammento
     * @param author autore
     * @param type e tipo
     */
    public Spectrum getSpectrum(String uri, String xPath, String author, String type);
    /**
     * restituisce il blocco di annotazioni raffinate riferite ad un frammento (autore e tipo sono facoltativi)
     * @return ritorna uno spetto
     * @param uri uri del frammento
     * @param xPath xpath del frammento
     * @param author autore
     * @param type e tipo
     */
    public Spectrum getXSpectrum(String uri, String xPath, String author, String type);
    /**
     * updates the given Spectrum
     * @param spectrum Spectrum to update
     * @throws Exception in case of DB errors
     * @return the number of row inserted
     */
    public int updateSpectrum  (Spectrum spectrum) throws Exception ;
    /**
     * Delete an annotation block
     * @return < 0 for errors; > 0 for succesful operation
     * @param uri uri for the annotations
     * @param xPath xpath for the annotations
     * @param author author of the annotation (or an empty String)
     * @param type type of the Annotation (or an Empty String)
     */
    public int deleteSpectrum(String uri, String xPath, String author, String type) ;
    /**
     * Deleete an (expanded) annotation block
     * @return < 0 for errors; > 0 for succesful operation
     * @param uri uri for the annotations
     * @param xPath xpath for the annotations
     * @param author author of the annotation (or an empty String)
     * @param type type of the Annotation (or an Empty String)
     */
    public int deleteXSpectrum(String uri, String xPath, String author, String type) ;
    /**
     * last database error
     * @return last error's descriprion
     */
    public String dbStatus() ;
    /**
     * restituisce  gli spettri in cui compaiono valori non nulli per i topic specificati
     * @param level il numero minimo di match per ogni topic restituito
     * @return restituisce un vettore di spettri che corrispondono ai parametri
     * @param topicURIs vettore di topic su cui fare la ricerca
     */
    public Vector searchXByTopic(Vector topicURIs, int level);
    
    /**
     * return the key of a given annotation author
     * @return null if the argument doesn'exist,else the key value
     * @param author the elelement to retrieve
     * @throws Exception in error case
     */
    public Integer selectAuthor(String author) throws Exception;
    
    /**
     * return the key of a given annotation type
     * @return null if the argument doesn'exist,else the key value
     * @param type the element type to retrieve
     * @throws Exception in error case
     */
    public Integer selectType(String type) throws Exception;
    /**
     * return the key of a given topic
     * @param topic tpoic to insert
     * @return null if the argument doesn'exist,else the key value
     * @throws Exception in error case
     */
    public Integer selectTopic(String topic) throws Exception;
    /**
     * inserts an annotation author into the database
     * @return null in error, Integer success
     * @param author the author to insert
     * @throws Exception in error case
     */
    public Integer insertAuthor(String author) throws Exception;
    /**
     * inserts an annotation type into the database
     * @return null in error, Integer success
     * @param anntype the annotation type to insert
     * @throws Exception in error case
     */
    public Integer insertType(String anntype) throws Exception;
    /**
     * inserts a topic into the database
     * @param topic tpoic to insert
     * @return null in error, Integer success
     * @throws Exception in error case
     */
    public Integer insertTopic(String topic) throws Exception;
    
    /**
     * delete a topic from the database, given the key
     * @return null in error, else 1
     * @param idauthor the id of the element to delete
     * @throws Exception in error
     */
    public Integer deleteAuthor(Integer idauthor) throws Exception;
    /**
     * delete a topic from the database, given the topicURI
     * @return null null in error, Integer (1) in success
     * @param author the element to delete
     * @throws Exception in error
     */
    public Integer deleteAuthor(String author) throws Exception;
    
    
    /**
     * delete a topic from the database, given the key
     * @return null in error, else 1
     * @param idtype the element to delete
     * @throws Exception in error
     */
    public Integer deleteType(Integer idtype) throws Exception;
    /**
     * delete a topic from the database, given the topicURI
     * @return null null in error, Integer (1) in success
     * @param anntype the element to delete
     * @throws Exception in error
     */
    public Integer deleteType(String anntype) throws Exception;
    
    /**
     * delete a topic from the database, given the key
     * @param idtopic key of the elemet do delete
     * @return null in error, else 1
     * @throws Exception in error
     */
    public Integer deleteTopic(Integer idtopic) throws Exception;
    /**
     * delete a topic from the database, given the topicURI
     * @param topic elemet to delete
     * @return null null in error, Integer (1) in success
     * @throws Exception in error
     */
    public Integer deleteTopic(String topic) throws Exception;
    /**
     * Per poter implementare le politiche di autogestione
     * della copertura concettuale delle annotazioni
     * occorre introdurre un metodo getTopicsCoverage
     * @return  TopicCoverage il risultato e` un vettore di coppie
     * topicURI/frequenza, dove la frequenza ? il numero di
     * annotazioni memorizzate riferite ad un dato topic,
     * ordinato in modo crescente (prima le frequenze piu` basse
     * e cosi` via...).
     * Viene restituito null in caso di errore
     * @throws Exception In caso di errore spara una eccezione
     */
    public TopicCoverage getTopicsCoverage() throws Exception;
    /**
     * Elimina un Synset, dato topicUri e linguaggio nella tabella sysnset
     * @return null in caso di errore altrimenti il valore della chiave in caso di successo
     * @param i chiave da eliminare
     * @throws Exception eccezione in caso di errore
     */
    public Integer deleteSynset(Integer i) throws Exception;
    /**
     * Elimina un Synset, dato topicUri e linguaggio nella tabella sysnset
     * @return null in caso di errore altrimenti il valore della chiave in caso di successo
     * @param topicURI l'uri a cui si riferisce il synset
     * @param lang lingua del synset
     * @throws Exception si` spara un'eccezione
     */
    public Integer deleteSynset(String topicURI, String lang) throws Exception;
    /**
     * elimina un synset, data la chiave
     * @return null in caso di errore
     * @param S da eliminare (ATTENZIONE: non elimina solo quello che
     *         si trova nelle occour ....)
     * @throws Exception In caso di problemi al DB spara un'eccezione
     */
    public Integer deleteSynset(Synset S) throws Exception;
    /**
     * Inserisce una riga nella tabella sysnset
     * @param topicURI l'uri a cui si riferisce il synset
     * @param lang lingua del synset
     * @return null in caso di errore Integer (il valore della chiave della tabella) in caso di successo
     * @throws Exception in caso di errori...
     */
    public Integer insertSynset(String topicURI, String lang) throws Exception;
    /**
     *Cerca una riga nella tabella sysnset
     *@param topicURI l'uri a cui si riferisce il synset
     *@param lang lingua del synset
     *@return null in caso di errore altrimenti il valore della chiave in caso di successo
     *@throws Exception for db errors
     */
    public Integer selectSynset(String topicURI, String lang)throws Exception;
    /**
     * loads a new synset, or add some mwords to an existing synset. If
     * an mword is already present, the operation will fail. updateSynset,
     * does not fail (but is slower).
     *
     * @param S Synset to insert/extend
     * @throws Exception for db problems/mword already present
     * @return null, if there is some error
     */
    public Integer loadSynset(Synset S) throws Exception;
    /**
     * uptates the synset values
     *
     * @return null for error, else an Integer
     * @param addMode false for substitution of frequency, true for adding
     * @param S Synset to insert/update
     * @throws Exception for database errors
     */
    public Integer updateSynset(Synset S, boolean addMode) throws Exception;
    /**
     * counts the total numer of inserted annotations
     * @throws Exception in case of database error
     * @return the number of annotation, a negative number in case of
     *         error
     */
    public int countAnnotations() throws Exception;
    /**
     * counts the total number of annotations given the topic they refer to
     *
     * @param  topicURI the topic to check
     * @throws Exception in case of seroius DB problems
     * @return the number of annotation, a negative number in case of
     *         error
     */
    public int countAnnotations(String topicURI) throws Exception;
}
