package it.polito.elite.doseWS.wrappers;

import com.hp.hpl.jena.ontology.*;
import com.hp.hpl.jena.vocabulary.*;
import com.hp.hpl.jena.rdf.model.*;
import com.hp.hpl.jena.util.iterator.ExtendedIterator;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.util.*;
import java.net.URL;
import java.net.URLConnection;

/** The Ontology wrapper.
 * @author <a href="mailto:federico.pescarmona@polito.it">Federico Pescarmona</a>
 */
public class OntologyImpl implements Ontology {
	private OntModel model;
	private final static String defaultOntoFilename = "http://localhost:8080/axis/ontology.owl";
	private String inputOntoFilename;
	private Vector allClasses;
	private Vector roots;
	private Hashtable relations;

	/** Constructor.
	 * @param inputOntoFileName The filename for the &quot;.rdf&quot; file into which the ontology is stored.
	 */
	public void OntologyImplFile(String inputOntoFilename) {
		try {
				this.inputOntoFilename = inputOntoFilename;
				InputStream in = Ontology.class.getClassLoader()
				.getResourceAsStream(inputOntoFilename);
				if (in == null) 
				{
					throw new IllegalArgumentException("File: " + inputOntoFilename
						+ " not found");
				}
				OntologyImplCommon(new InputStreamReader(in));
				} 
		catch (Exception e) 
		{
			System.out.println("Failed: " + e.getMessage());
		}
	}
	public void OntologyImplCommon(InputStreamReader in)
	{
		try
		{
			model = ModelFactory.createOntologyModel();
			model.read(in, null);
		}
		catch (Exception e) 
		{
			System.out.println("Failed: " + e.getMessage());
		}
	}

	/** Constructor. Loads the ontology from a default file. */
	public OntologyImpl() {
		OntologyImplURL(defaultOntoFilename);
	}
	
	public OntologyImpl(String resource,boolean isURI)
	{
		if(isURI)
			OntologyImplURL(resource);
		else
			OntologyImplFile(resource);
	}
	
	public void OntologyImplURL(String url)
	{
		try
		{
			URL ontoURL = new URL(url);
			URLConnection conn = ontoURL.openConnection();
			InputStreamReader in = new InputStreamReader(conn.getInputStream());
			if (in == null) {
				throw new IllegalArgumentException("URL: " + url
						+ " not found");
			}
			OntologyImplCommon(in);
		}
		catch(Exception e)
		{
			System.out.println("Failed: " + e.getMessage());
		}
		
	}

	/** Prints the ontology concepts on the output console, formatted as tuples. */
	public void printTuples() { // scrive tuple su schermo

		try {
			StmtIterator iter = model.listStatements();
			while (iter.hasNext()) {
				Statement stmt = iter.nextStatement(); // get next statement
				Resource subject = stmt.getSubject(); // get the subject
				Property predicate = stmt.getPredicate(); // get the predicate
				RDFNode object = stmt.getObject(); // get the object

				System.out.print(subject.toString());
				System.out.print(" " + predicate.toString() + " ");
				if (object instanceof Resource) {
					System.out.print(object.toString());
				} else {
					System.out.print(" \"" + object.toString() + "\"");
				}
				System.out.println(" .");
			}
		} catch (RDFException e) {
			System.out.println("Failed: " + e.getMessage());
		}
	}

	/** Prints the ontology concepts on the output console formatted as valid XML. */
	public void printXML() {
		try {
			model.write(new PrintWriter(System.out)); // scrive su schermo in formato XML
		} catch (RDFException e) {
			System.out.println("Failed: " + e.getMessage());
		}
	}

	/**
	 * Provides a Vector containing all ontology concepts as URIs.
	 * @return The Vector containing all ontology concepts.
	 */
	public Vector getAllClasses() {
		Iterator iter;
		if (allClasses == null) {
			iter = model.listClasses();
			allClasses = new Vector();
			OntClass c;
			while(iter.hasNext()) {
				c = (OntClass) iter.next();
				allClasses.add(c.getURI());
			}
		}
		return allClasses;
	}
	
	public Set getRoots() {
		if (roots == null) {
			roots = new Vector();
			ExtendedIterator iter = model.listClasses();
			OntClass node;
			while(iter.hasNext()) {
				node = (OntClass) iter.next();
				if((!node.hasProperty(RDFS.subClassOf))||(node.hasProperty(RDFS.subClassOf)&&(node.getSuperClass().equals(node))))
						roots.add(node.getURI());
			}
		}
		return new HashSet(roots);
	}

	/** Retrieved the label of an ontology concept, given the URI.
	 * @param uri The concept URI.
	 * @return The concept label.
	 */
	public String getLabel(String uri) {
		String label = null;
		try {
			OntClass class1 = model.getOntClass(uri);
			label = class1.getLabel(null);
		} catch (Exception e) {
		}
		return label;
	}
	
	/**
	 * Returns the Ontology class.
	 * @param uri The URI of the class.
	 * @return The class.
	 */
	public OntClass getOntClass(String uri) {
		try {
			return model.getOntClass(uri);
		} catch (Exception e) {
			System.out.println("Exception: " + e);
			return null;
		}
	}

	/** Retrieves all direct sub-classes of the given concept.
	 * @param uri The URI of the &quot;father&quot; concept.
	 * @return An Iterator containing all sub-classes expressed as URIs.
	 */
	public Iterator getSubClasses(String uri) {
		Iterator iter = null;
		try {
			Vector v = new Vector();
			OntClass class1 = model.getOntClass(uri);
			Iterator iter2 = (Iterator) class1.listSubClasses(true);
			while(iter2.hasNext()) {
				v.add(((OntClass) iter2.next()).getURI());
			}
			iter = v.iterator();
		} catch (Exception e) {
			System.err.println("Error in getSonsAsUri() for uri " + uri + ": "+ e);
		}
		return iter;
	}

	/** Retrieves all direct super-classes of the given concept.
	 * @param uri The URI of the concept.
	 * @return An Iterator containing all super-classes expressed as URIs.
	 */
	public Iterator getSuperClasses(String uri) {
		Iterator iter = null;
		try {
			Vector v = new Vector();
			OntClass class1 = model.getOntClass(uri);
			Iterator iter2 = (Iterator) class1.listSuperClasses(true);
			while(iter2.hasNext()) {
				v.add(((OntClass) iter2.next()).getURI());
			}
			iter = v.iterator();
		} catch (Exception e) {
			System.err.println("Ontology: error getting fathers of " + uri
					+ " " + e);
		}
		return iter;
	}

	/** Counts the number of sons for a given concept.
	 * @param uri The concept label.
	 * @return The number of sons.
	 */
	public int getNumberOfSubClasses(String uri) {
		int k = 0;
		try {
			OntClass c = model.getOntClass(uri);
			ExtendedIterator iter = c.listSubClasses(true);
			while (iter.hasNext()) {
				iter.next();
				k++;
			}
		} catch (Exception e) {
			System.err.println("Ontology: error getting number of sons of "
					+ uri + " " + e);
		}
		return k;
	}
	
	/**
	 * Given a concept and a relation, returns an Iterator over the classes the source concept is related to (through the given relation).
	 * @param nodeURI The source concept as URI.
	 * @param relationURI The relation as URI.
	 * @return An Iterator over the URIs of the related concepts.
	 */
	public Vector getRelationEnds(String nodeURI, String relationURI) {
		Vector ends = new Vector();
		Property prop = model.getProperty(relationURI);
		OntClass node = model.getOntClass(nodeURI);
		StmtIterator iter = node.listProperties(prop);
		while (iter.hasNext()) {
			Statement stmt = iter.nextStatement();
			Resource topic = stmt.getResource();
			ends.add(topic.getURI());
		}
		return ends;
	}

	/**
	 * Returns all relations of the ontology with their weights.
	 * @return The relations mapped as URI, weight.
	 */
	public Hashtable getAllRelations() {
		if (relations == null) {
			relations = new Hashtable();
			ExtendedIterator opIter = model.listObjectProperties();
			ObjectProperty op;
			String ns;
			AnnotationProperty weight;
			double w;
			while(opIter.hasNext()) {
				op = (ObjectProperty) opIter.next();
				ns = op.getNameSpace();
				weight = model.getAnnotationProperty(ns + "weight");
				w = op.getProperty(weight).getFloat();
				relations.put(op.getURI(),new Double(w));
			}
		}
		return relations;
	}

	/**
	 * Returns the language-dependent title of a concept.
	 * @param topicURI The URI of a concept.
	 * @param language The language code.
	 * @return The title.
	 */
	public String getTitle(String topicURI, String language) {
		String title = "";
		String ns;
		OntClass node = model.getOntClass(topicURI);
		ns = node.getNameSpace();
		AnnotationProperty titleP = model.getAnnotationProperty(ns + "title");
		StmtIterator iter = node.listProperties(titleP);
		Statement stmt;
		while(iter.hasNext()) {
			stmt = iter.nextStatement();
			if (stmt.getLanguage().equalsIgnoreCase(language)) {
				title = stmt.getString();
				break;
			}
		}
		return title;
	}

	/**
	 * Returns the language-dependent definition of a concept.
	 * @param topicURI The URI of a concept.
	 * @param language The language code.
	 * @return The definition.
	 */	
	public String getDefinition(String topicURI, String language) {
		String def = "";
		String ns;
		OntClass node = model.getOntClass(topicURI);
		ns = node.getNameSpace();
		AnnotationProperty defP = model.getAnnotationProperty(ns + "definition");
		StmtIterator iter = node.listProperties(defP);
		Statement stmt;
		while(iter.hasNext()) {
			stmt = iter.nextStatement();
			if (stmt.getLanguage().equalsIgnoreCase(language)) {
				def = stmt.getString();
				break;
			}
		}
		return def;
	}
}